/*
 * searchservice process.rs
 * - route incoming nats requests to their specific functions
 *
 * Copyright (C) 2025 Real Microsoft, LLC
 *
 * This program is free software: you can redistribute it and/or modify it under the terms of the GNU Affero General Public License as published by the Free Software Foundation, version 3.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more details.
 *
 * You should have received a copy of the GNU Affero General Public License along with this program. If not, see <https://www.gnu.org/licenses/>.
*/

use std::sync::Arc;
use async_nats::{jetstream, Subject};
use log::{debug, error, warn};
use tokio::sync::Mutex;
use futures::StreamExt;
use asklyphe_common::ldb::{sitestore, DBConn};
use asklyphe_common::nats::searchservice;
use asklyphe_common::nats::searchservice::{SearchSrvcQuery, SearchSrvcRequest, SearchSrvcResponse, SearchRequest, SiteCountResponse, SearchResponse};
use crate::{algorithm, hacks};

pub async fn process(query: SearchSrvcQuery, db: DBConn) -> SearchSrvcResponse {
    // a future is used so that the whole program doesn't die if an algorithm panics
    let response = tokio::spawn(async move {
        match query.request {
            SearchSrvcRequest::SearchRequest(req) => {
                search_request(req, &db).await
            }
            SearchSrvcRequest::SiteCountRequest => {
                count_websites(&db).await
            }
        }}).await;

    if let Ok(response) = response {
        response
    } else if let Err(e) = response {
        SearchSrvcResponse::OtherError(e.to_string())
    } else {
        unreachable!()
    }
}

pub async fn search_request(req: SearchRequest, db: &DBConn) -> SearchSrvcResponse {
    let words_initial: Vec<String> = req.query.split_whitespace().map(|s| s.to_string()).collect();
    let mut words = vec![];
    let mut no_separator_flag = false;
    for word in words_initial {
        if hacks::is_from_language_that_doesnt_use_word_separators(&word) {
            words.extend(word.chars().map(|c| c.to_string()).collect::<Vec<String>>());
            no_separator_flag = true;
        } else {
            words.push(word);
        }
    }

    match algorithm::search(&db, words, no_separator_flag).await {
        Some(results) => {
            SearchSrvcResponse::SearchResponse(results)
        }
        None => {
            SearchSrvcResponse::SearchResponse(SearchResponse {
                results: vec![],
                blocked: vec![],
                pagerank_time_seconds: 0.0,
                total_query_seconds: 0.0,
                max_relevance: 0.0,
                exact_phrase_found: false,
            })
        }
    }
}

pub async fn count_websites(db: &DBConn) -> SearchSrvcResponse {
    if let Ok(count) = sitestore::count_websites(db).await {
        SearchSrvcResponse::SiteCountResponse(SiteCountResponse {
            count,
        })
    } else {
        SearchSrvcResponse::OtherError("couldn't retrieve count ):".to_string())
    }
}